﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data;
using System.IO;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using KMeansProject;
using System.ComponentModel;

namespace ArticleClassifier
{
    public partial class Form1 : Form
    {
        Dictionary<string, int> wordFrequencyDictionary;

        List<TrainItem> trainingDataSet;
        string[] files;

        private KMeans objKMeans;
        private BackgroundWorker objBackgroundWorker;
        private KMeansEventArgs kmeansEA;
        private EuclideanDistance objDistance;
        public Form1()
        {
            InitializeComponent();
        }

        private void btnLoadTrainingSet_Click(object sender, EventArgs e)
        {
            using (FolderBrowserDialog fbd = new FolderBrowserDialog())
            {
                fbd.SelectedPath = @"C:\Users\Vancho\Desktop\YouTube\C#\3. Article Classification KMeans";
                if (fbd.ShowDialog() == DialogResult.OK)
                {
                    files = Directory.GetFiles(fbd.SelectedPath);

                    trainingDataSet = new List<TrainItem>();
                    foreach(string file in files)
                    {
                        string fileContent = File.ReadAllText(file);
                        string processedFile = "";
                        if (checkBox1.Checked)
                            processedFile = Regex.Replace(fileContent, "[^a-zA-Z ]", "");
                        else
                            processedFile = fileContent;

                        trainingDataSet.Add(new TrainItem() { FilePath = file, InputVector = null, Text = processedFile });

                    }

                    numericUpDown1.Minimum = 1;
                    numericUpDown1.Maximum = trainingDataSet.Count;
                }
            }
        }

        private void btnPreview_Click(object sender, EventArgs e)
        {
            txtFile.Text = trainingDataSet[(int)numericUpDown1.Value - 1].Text; 
        }

        private void btnWord2Vec_Click(object sender, EventArgs e)
        {
            wordFrequencyDictionary = new Dictionary<string, int>();

            foreach(TrainItem item in trainingDataSet)
            {
                string[] words = item.Text.Split(' ');
                foreach(string word in words)
                {
                    string wordLowerCaps = word.ToLower();
                    if (wordFrequencyDictionary.ContainsKey(wordLowerCaps))
                    {
                        wordFrequencyDictionary[wordLowerCaps] += 1;
                    }
                    else
                    {
                        wordFrequencyDictionary[wordLowerCaps] = 1;
                    }
                }
            }

            wordFrequencyDictionary = (from entry in wordFrequencyDictionary orderby entry.Value descending select entry).ToDictionary(pair => pair.Key, pair => pair.Value);

            lblWords.Text = "Words: " + wordFrequencyDictionary.Count;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            frmWords frmWords = new frmWords(wordFrequencyDictionary);
            frmWords.OnApplyChanges += FrmWords_OnApplyChanges;
            frmWords.Show();
        }

        private void FrmWords_OnApplyChanges(object sender, Dictionary<string,int> e)
        {
            wordFrequencyDictionary = e;

            lblWords.Text = "Words: " + wordFrequencyDictionary.Count;
        }

        private void btnTrain_Click(object sender, EventArgs e)
        {
            objDistance = new EuclideanDistance();
            int clusters = (int)nClusters.Value;
            objKMeans = new KMeans(clusters, new EuclideanDistance());
            
            objBackgroundWorker = new BackgroundWorker();
            
            int itemIndex = 0;
            foreach(TrainItem item in trainingDataSet)
            {
                Dictionary<string, int> localFrequency = new Dictionary<string, int>();
                string[] words = item.Text.Split(' ');

                foreach(string word in words)
                {
                    string lowerWord = word.ToLower();
                    if (!localFrequency.ContainsKey(lowerWord))
                        localFrequency[lowerWord] = 1;
                    
                }

                double[] inputVector = new double[wordFrequencyDictionary.Count];

                foreach(KeyValuePair<string,int> pair in localFrequency)
                {
                    if (wordFrequencyDictionary.ContainsKey(pair.Key))
                    {
                        int index = wordFrequencyDictionary.Keys.ToList().IndexOf(pair.Key);
                        inputVector[index] = pair.Value;
                    }
                }

                trainingDataSet[itemIndex].InputVector = inputVector;
                itemIndex++;
            }

            objBackgroundWorker.WorkerReportsProgress = true;
            objBackgroundWorker.DoWork += ObjBackgroundWorker_DoWork;
            objBackgroundWorker.RunWorkerCompleted += ObjBackgroundWorker_RunWorkerCompleted;
            objBackgroundWorker.ProgressChanged += ObjBackgroundWorker_ProgressChanged;

            List<double[]> inputTrainingSet = new List<double[]>();
            foreach (TrainItem trainItem in trainingDataSet)
                inputTrainingSet.Add(trainItem.InputVector);

            objBackgroundWorker.RunWorkerAsync(inputTrainingSet.ToArray());
        }

        private void ObjBackgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            double[][] inputDataset = e.Argument as double[][];
            
            objKMeans.UpdateProgress += (x, y) => {
                objBackgroundWorker.ReportProgress(0, y);
            };
            e.Result = objKMeans.Run(inputDataset);
        }
        private void ObjBackgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            Centroid[] centroids = e.Result as Centroid[];
            Console.WriteLine("Work is done!");
            button1.Enabled = true;
        }

        private void ObjBackgroundWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            System.Console.WriteLine(">>> Progress Changed");
            kmeansEA = e.UserState as KMeansEventArgs;
            if (kmeansEA != null)
            {
                foreach (Centroid centroid in kmeansEA.CentroidList)
                {
                    System.Console.WriteLine("Centroid: ");
                }
            }
        }

        private void btnVisualize_Click(object sender, EventArgs e)
        {
            frmVisualization objVisualizations = new frmVisualization();

            using (OpenFileDialog ofd = new OpenFileDialog())
            {
                if (ofd.ShowDialog()== DialogResult.OK)
                {
                    string fileContent = File.ReadAllText(ofd.FileName);
                    string processedContent = "";
                    if (checkBox1.Checked)
                        processedContent = Regex.Replace(fileContent, "[^0-9a-zA-Z ]", "");
                    else
                        processedContent = fileContent;

                    Dictionary<string, int> localFrequency = new Dictionary<string, int>();
                    string[] words = processedContent.Split(' ');

                    foreach (string word in words)
                    {
                        string lowerWord = word.ToLower();
                        if (!localFrequency.ContainsKey(lowerWord))
                            localFrequency[lowerWord] = 1;
                    }

                    double[] inputVector = new double[wordFrequencyDictionary.Count];

                    foreach (KeyValuePair<string, int> pair in localFrequency)
                    {
                        if (wordFrequencyDictionary.ContainsKey(pair.Key))
                        {
                            int index = wordFrequencyDictionary.Keys.ToList().IndexOf(pair.Key);
                            inputVector[index] = pair.Value;
                        }
                    }

                    int itemClass = objKMeans.Classify(inputVector);
                    MessageBox.Show("Belongs to: " + itemClass + " class");
                }
            }
        }

        private void objPresentResults_Click(object sender, EventArgs e)
        {
            List<int> uniqueLabels = new List<int>();
            foreach(TrainItem item in trainingDataSet)
            {
                int resultClass = objKMeans.Classify(item.InputVector);
                item.resultClass = resultClass;

                if (!uniqueLabels.Contains(resultClass))
                    uniqueLabels.Add(resultClass);
            }

            frmVisualization objVisualizations = new frmVisualization();

            foreach (int label in uniqueLabels)
            {
                List<ColumnItem> columnItems = new List<ColumnItem>();
                foreach(TrainItem item in trainingDataSet)
                {
                    if (item.resultClass == label)
                        columnItems.Add(new ColumnItem() { Name = Path.GetFileName(item.FilePath), Text = item.Text });
                }
                objVisualizations.AddItem(String.Format("Cluster: {0}", label), columnItems);
            }

            objVisualizations.Show();
        }
    }
}
